from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass, BaseClass,init_class_kwargs,get_object_model_class

@dataclass
class GendersFields(BaseFieldsClass):
    text = "text"
    gendersRangeType = "gendersRangeType"
    criterionId = "criterionId"
    criterionType = "criterionType"
    type = "type"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@dataclass(init=False)
class Genders(BaseClass):
    text: str
    gendersRangeType: str
    criterionId: int
    criterionType: str
    type: str
    def __init__(self, ages=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, ages, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "text": self.text,
            "gendersRangeType": self.gendersRangeType,
            "criterionId": self.criterionId,
            "criterionType": self.criterionType,
            "type": self.type
        })
    @staticmethod
    def from_dict(obj: Any) -> 'Genders':
        cls_object, keys = get_object_model_class(obj, Genders, None);
        return Genders(cls_object)


from ....base.firebase_collection import FirebaseCollectionBase
from .....utils import ArrayUtils, DictUtils, Crypto, get_object_model_class, init_class_kwargs
from .client_fields import ClientFields, ClientFieldProps, STANDARD_FIELDS, CLIENT_COLLECTION
from .....utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from werkzeug.security import gen_salt
import time
from typing import Any
from dataclasses import dataclass
import json
from authlib.oauth2.rfc6749 import ClientMixin
from authlib.oauth2.rfc6749.util import list_to_scope, scope_to_list
import secrets
import pydash

ADAFRI_SCOPES = ["openid", "profile", "email", "profile", "profile.read", "profile.update", "accounts.read", "accounts.update", "accounts_management"]

@dataclass(init=False)
class OAuthClient(ClientMixin, FirebaseCollectionBase):
    id: str
    name: str
    uid: str
    description: str
    client_id: str
    client_secret: str
    uri: str
    grant_types: list[str]
    response_types: list[str]
    token_endpoint_auth_method: str
    redirect_uris: list[str]
    scopes: list[str]
    scope: str
    client_id_issued_at: int
    createdAt: any
    allowed_redirect_uris: list[str]
    default_redirect_uri: str
    allow_direct_grant_access: bool
    token_endpoint_auth_methods: list[str]
    
    def __init__(self, client=None, default_redirect_uri=None, **kwargs):
        if type(client) is str:
            client = {"client_id": client} 
        (cls_object, keys, data_args) = init_class_kwargs(self, client, STANDARD_FIELDS, ClientFieldProps, CLIENT_COLLECTION, ['id'], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            setattr(self, key, cls_object[key])
        if getattr(self, 'allowed_redirect_uris',None) is None or bool(self.allowed_redirect_uris) is False:
            if getattr(self, 'redirect_uris',None) is not None:
                self.allowed_redirect_uris = self.redirect_uris
        #self.default_redirect_uri = default_redirect_uri


    @staticmethod
    def generate_model(_key_="default_value"):
        user = {};
        props = ClientFieldProps
        for k in DictUtils.get_keys(props):
            user[k] = props[k][_key_];
        return user;

    @staticmethod
    def from_dict(client = None, db=None, collection_name=None, default_redirect_uri=None) -> 'OAuthClient':
        cls_object, keys = get_object_model_class(client, OAuthClient, ClientFieldProps);
        _client = OAuthClient(cls_object, db=db, collection_name=collection_name, default_redirect_uri=default_redirect_uri)
        return _client

    def query(self, query_params: list, first=False, limit=None):
        result = [];
        query_result = self.custom_query(query_params, first=first, limit=limit)
        if bool(query_result):
            if first:
                return OAuthClient.from_dict(client=query_result, db=self.db, collection_name=self.collection_name, default_redirect_uri=getattr(self,  'default_redirect_uri', None))
            else:
                for doc in query_result:
                    result.append(OAuthClient.from_dict(client=doc, db=self.db, collection_name=self.collection_name, default_redirect_uri=getattr(self,  'default_redirect_uri', None)))
                return result
        if first:
                return None
        return [];

    def get_by_client_id(self, id) -> 'OAuthClient':
        if id is None or bool(id) is False:
            return None
        return self.query([{"key": "client_id", "comp": "==", "value": id}], True)
       

    def getOAuthClient(self) -> 'OAuthClient':
        if bool(self.id):
            doc = self.document_reference().get();
            if doc.exists is False:
                return None;
            return OAuthClient.from_dict(doc.to_dict(), db=self.db, collection_name=self.collection_name);
        if bool(self.client_id):
            return self.query([{"key": ClientFields.client_id, "comp": "==", "value": self.client_id}], True)

    @staticmethod
    def generate(**kwargs) -> 'ApiResponse':
        data_dict = DictUtils.pick_fields(kwargs, ClientFields.filtered_keys('mutable', True));
        client_model = OAuthClient.from_dict(DictUtils.merge_dict(data_dict, OAuthClient.generate_model()));
        
        if bool(client_model.to_json()) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Empty request","INVALID_REQUEST", 1)).to_json()
        
        if bool(client_model.name) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("name required","INVALID_REQUEST", 1));
        
        if bool(client_model.uid) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("uid required","INVALID_REQUEST", 1));

        client_model.client_id = gen_salt(24)
        client_model.id = Crypto().generate_id(client_model.name+"~"+client_model.uid+"~"+client_model.client_id);
        client_model.client_id_issued_at = int(time.time())
        
        if client_model.token_endpoint_auth_method == 'none':
            client_model.client_secret = ''
        else:
            client_model.client_secret = gen_salt(48)
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, client_model, None);
    
    def update(self, data):
        try:
            last_value = self.to_json();
            filtered_value = pydash.pick(data, ClientFields.filtered_keys('editable', True));
            new_value = DictUtils.merge_dict(filtered_value, self.to_json());
            changed_fields = DictUtils.get_changed_field(last_value, new_value);
            data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
            if bool(data_update) is False:
                return None;
            self.document_reference().set(data_update, merge=True)
            return DictUtils.dict_from_keys(self.getOAuthClient().to_json(), changed_fields);
        except Exception as e:
            print(e)
            return None;

    def get_default_redirect_uri(self):
        return self.redirect_uris[0]
    
    def check_response_type(self, response_type):
        return response_type in self.response_types
    
    def check_client_secret(self, client_secret):
        return secrets.compare_digest(self.client_secret, client_secret)
    
    def check_endpoint_auth_method(self, method, endpoint):
        if endpoint == 'token':
            # if client table has ``token_endpoint_auth_method``
            token_endpoint_auth_methods = getattr(self, 'token_endpoint_auth_methods', [])
            return self.token_endpoint_auth_method == method or method in token_endpoint_auth_methods
        return True
    
    def check_grant_type(self, grant_type):
        return grant_type in self.grant_types
    
    def get_allowed_scope(self, scope):
        if not scope:
            return ''
        allowed = set(scope_to_list(self.scope))
        return list_to_scope([s for s in scope.split() if s in allowed])
    
    def check_redirect_uri(self, redirect_uri):
        return redirect_uri in self.allowed_redirect_uris

    def get_client_id(self):
        return self.client_id
    


from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass, BaseClass, init_class_kwargs,get_object_model_class

@dataclass
class AdsSchedulesFields(BaseFieldsClass):
    id = "id"
    dayEN = "dayEN"
    dayFR = "dayFR"
    endHour = "endHour"
    endMinute = "endMinute"
    startHour = "startHour"
    startMinute = "startMinute"
    end_hour_text = "end_hour_text"
    start_hour_text = "start_hour_text"
    criterionId = "criterionId"
    criterionType = "criterionType"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@dataclass(init=False)
class AdsSchedules(BaseClass):
    id: str
    dayEN: str
    dayFR: str
    endHour: str
    endMinute: str
    startHour: str
    startMinute: str
    start_hour_text: str
    end_hour_text: str
    criterionId: int
    criterionType: str
    def __init__(self, schedules=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, schedules, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "dayEn": self.dayEN,
            "endHour": self.endHour,
            "endMinute": self.endMinute,
            "startHour": self.startHour,
            "startMinute": self.startMinute,
            "start_hour_text": self.start_hour_text,
            "end_hour_text": self.end_hour_text,
            "criterionId": self.criterionId,
            "criterionType": self.criterionType,
        })
    @staticmethod
    def from_dict(obj: Any) -> 'AdsSchedules':
        cls_object, keys = get_object_model_class(obj, AdsSchedules, None);
        return AdsSchedules(cls_object)
        


from dataclasses import dataclass
from  ....utils.utils import DictUtils, BaseFieldsClass 
import os

ACCOUNT_COLLECTION = os.environ.get('ACCOUNT_COLLECTION');
ACCOUNT_LINK_COLLECTION = os.environ.get('ACCOUNT_LINK_COLLECTION');

@dataclass
class AccountFields(BaseFieldsClass):
    id = "id"
    aacid = "aacid"
    account_value = "account_value"
    creationDate = "creationDate"
    creationDateString = "creationDateString"
    name = "name"
    owner = "owner"
    owner_email = "owner_email"
    status = "status"
    totalClics = "totalClics"
    totalCosts = "totalCosts"
    totalImpressions = "totalImpressions"
    usedPackTest = "usedPackTest"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

@dataclass
class AccountLinkFields(BaseFieldsClass):
    id = "id"
    linkDate = "linkDate"
    linkDateString = "linkDateString"
    owner = "owner"
    owner_email = "owner_email"
    target = "target"
    target_email = "target_email"
    status = "status"
    linked_account = "linked_account"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)



AccountFieldProps = {
    AccountFields.id: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountFields.aacid: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountFields.account_value: {
        "type": (int or float),
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": 0
    },
    AccountFields.creationDateString: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": False,
        "pickable": True,
        "default_value": ""
    },
    AccountFields.creationDate: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountFields.name: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountFields.owner: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountFields.owner_email: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountFields.status: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountFields.totalClics: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountFields.totalCosts: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountFields.totalImpressions: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountFields.usedPackTest: {
        "type": bool,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": False
    }
}
AccountLinkFieldProps = {
    AccountLinkFields.id: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountLinkFields.linkDate: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountLinkFields.linkDateString: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountLinkFields.owner: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountLinkFields.owner_email: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountLinkFields.target: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountLinkFields.target_email: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountLinkFields.status: {
        "type": str,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": ""
    },
    AccountLinkFields.linked_account: {
        "type": dict,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "pickable": True,
        "default_value": None
    }
}

STANDARD_FIELDS = AccountFields(fields_props=AccountFieldProps).filtered_keys('pickable', True)
ACCOUNT_PICKABLE_FIELDS = AccountFields(fields_props=AccountFieldProps).filtered_keys('pickable', True)
LINK_STANDARD_FIELDS = AccountLinkFields(fields_props=AccountLinkFieldProps).filtered_keys('pickable', True)



from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass, BaseClass,init_class_kwargs,get_object_model_class

@dataclass
class CustomParametersFields(BaseFieldsClass):
    key = "key"
    value = "value"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@dataclass(init=False)
class CustomParameters(BaseClass):
    key: str
    value: str
    def __init__(self, custom_parameters=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, custom_parameters, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "key": self.key,
            "value": self.value
        })
    @staticmethod
    def from_dict(obj: Any) -> 'CustomParameters':
        cls_object, keys = get_object_model_class(obj, CustomParameters, None);
        return CustomParameters(cls_object)


import pycountry
import gettext

class Country(object):
    def __init__(self, locales=['fr']):
        self.locales = locales;
    
    def get_countries(self):
        lang =  gettext.translation('iso3166-1', pycountry.LOCALES_DIR, languages=self.locales);
        lang.install()
        countries_list = list(pycountry.countries)
        countries = []
        for c in countries_list:
            country = self.get_country(country=c)
            countries.append(country)
        return countries;
    
    def get_country(self, country=None, country_code=None):
        c = country;
        data_country = None;
        if country is None:
            if country_code is None:
                return data_country;
        
            c = pycountry.countries.lookup(country_code);
        
        if c is not None:
            name = getattr(c, 'name', '')
            data_country = {
                "name": _(name),
                "official_name": _(getattr(c, 'official_name', name)),
                "numeric": getattr(c, 'numeric', ''),
                "alpha_2": getattr(c, 'alpha_2', None),
                "alpha_3": getattr(c, 'alpha_3', None),
                'flag': getattr(c, 'flag', None),
            }
        return data_country


from .models import (account, account_fields)
from .models.account import Account, AccountLink, filter_account_request_fields
from .models.account_fields import (AccountFields, AccountFieldProps, AccountLinkFields, AccountLinkFieldProps)



from dataclasses import dataclass
from  ....utils.utils import DictUtils, BaseFieldsClass 
import os

USERS_COLLECTION = os.environ.get('USER_COLLECTION');
@dataclass
class PhoneInfosFields(BaseFieldsClass):
    countryCode = "countryCode";
    dialCode = "dialCode";
    e164Number = "e164Number";
    internationalNumber = "internationalNumber";
    nationalNumber = "nationalNumber";
    number = "number";
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.class_fields_props = PhoneInfosFieldProps

PhoneInfosFieldProps = {
    PhoneInfosFields.countryCode: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    PhoneInfosFields.dialCode: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    PhoneInfosFields.e164Number: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    PhoneInfosFields.nationalNumber: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    PhoneInfosFields.internationalNumber: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    PhoneInfosFields.number: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
}

@dataclass
class UserFields(BaseFieldsClass):
    uid = "uid";
    email = 'email';
    password = 'password';
    displayName = 'displayName';
    firstName = 'firstName';
    lastName = 'lastName';
    provider = 'provider';
    entrepriseName = 'entrepriseName';
    entrepriseUrl = 'entrepriseUrl';
    isConnectWithMailAndPassword = 'isConnectWithMailAndPassword';
    address = 'addresse';
    photoURL = 'photoURL';
    profileCompleted = 'profileCompleted';
    telephone = 'telephone';
    postalCode = 'postalCode';
    token = 'token';
    deviceInfo = 'deviceInfo';
    phoneInfo = 'phoneInfo';
    country = 'country';
    showPushToken = 'showPushToken'
    authorizedPush = 'authorizedPush';
    hasApprouvedPolicy = 'hasApprouvedPolicy';
    _emailValidationSendDate = 'emailValidationSendDate'
    status = 'status'
    businessType = 'businessType'
    user_type = 'user_type'
    phoneVerified = 'phoneVerified'
    emailVerified = 'emailVerified'
    postal = 'postal'
    _pwResetSendDate = '_pwResetSendDate'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.class_fields_props = UserFieldProps



UserFieldProps = {
    UserFields.uid: {
        "type": str,
        "required": True,
        "mutable": False,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    UserFields.email: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": False,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    UserFields.address: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": None,
        "pickable": True,
    },
    UserFields.password: {
        "type": str,
        "required": True,
        "mutable": True,
        "editable": True,
        "interactive": False,
        "default_value": "",
        "pickable": True,
    },
    UserFields.firstName: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    UserFields.lastName: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    UserFields.displayName: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    UserFields.postalCode: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    UserFields.telephone: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    UserFields.authorizedPush: {
        "type": bool,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": False,
        "pickable": True,
    },
    UserFields.hasApprouvedPolicy: {
        "type": bool,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": False,
        "pickable": True,
    },
    UserFields.profileCompleted: {
        "type": bool,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": False,
        "pickable": True,
    },
    UserFields.entrepriseUrl: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    UserFields.entrepriseName: {
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    UserFields.photoURL:{
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": True,
        "default_value": "",
        "pickable": True,
    },
    UserFields.provider:{
        "type": str,
        "required": False,
        "mutable": True,
        "editable": False,
        "interactive": False,
        "default_value": "https://adafri.app",
        "pickable": True,
    },
    UserFields.isConnectWithMailAndPassword:{
        "type": bool,
        "required": False,
        "mutable": False,
        "editable": False,
        "interactive": False,
        "default_value": False,
        "pickable": True,
    },
    UserFields.token:{
        "type": list[str],
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": False,
        "default_value": [],
        "pickable": True,
    },
    UserFields.phoneInfo:{
        "type": dict,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": False,
        "default_value": None,
        "pickable": True,
    },
    UserFields.deviceInfo:{
        "type": dict,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": False,
        "default_value": None,
        "pickable": True,
    },
    UserFields.status:{
    "type": str,
    "required": False,
    "mutable": True,
    "editable": True,
    "interactive": False,
    "default_value": "",
    "pickable": True,
    },
    UserFields._emailValidationSendDate:{
    "type": str,
    "required": False,
    "mutable": True,
    "editable": True,
    "interactive": False,
    "default_value": "",
    "pickable": True,
    },
    UserFields._pwResetSendDate:{
    "type": str,
    "required": False,
    "mutable": True,
    "editable": True,
    "interactive": False,
    "default_value": "",
    "pickable": True,
    },
UserFields.businessType:{
    "type": str,
    "required": False,
    "mutable": True,
    "editable": True,
    "interactive": False,
    "default_value": None,
    "pickable": True,
    },
    UserFields.user_type:{
    "type": str,
    "required": False,
    "mutable": True,
    "editable": True,
    "interactive": False,
    "default_value": "",
    "pickable": True,
    },
     UserFields.postal:{
        "type": str,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": False,
        "default_value": None,
        "pickable": True,
    },
    UserFields.phoneVerified:{
        "type": bool,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": False,
        "default_value": False,
        "pickable": True,
    },
    UserFields.emailVerified:{
        "type": bool,
        "required": False,
        "mutable": True,
        "editable": True,
        "interactive": False,
        "default_value": False,
        "pickable": True,
    }    
}

USER_FIELDS = UserFields(fields_props=UserFieldProps).filtered_keys('pickable', True)
PHONE_FIELDS = PhoneInfosFields(fields_props=PhoneInfosFieldProps).filtered_keys('pickable', True)



from dataclasses import dataclass
from typing import Any
from  ....utils.utils import ArrayUtils, DictUtils, BaseFieldsClass, BaseClass, init_class_kwargs,get_object_model_class
from ..data.adsChannel import ads_channel_data
@dataclass
class AdChannelFields(BaseFieldsClass):
    obj_id = "obj_id"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@dataclass(init=False)
class AdChannel(BaseClass):
    obj_id: str
    primary: dict
    secondary: dict
    def __init__(self, adChannel=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, adChannel, ['obj_id'], None, None, [], **kwargs)
        objective_value = ArrayUtils.find(ads_channel_data, lambda x: x['obj_id'] == cls_object['obj_id'])
        if objective_value is not None:
            for key in objective_value:
                setattr(self, key, objective_value[key]) 
        # for key in keys:
        #     setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "obj_id": self.obj_id,
            "primary": self.primary,
            "secondary": self.secondary
        })
    @staticmethod
    def from_dict(obj: Any) -> 'AdChannel':
        cls_object, keys = get_object_model_class(obj, AdChannel, None);
        return AdChannel(cls_object)


import requests
from authlib.jose import JsonWebToken, KeySet
from joserfc.jwk import RSAKey
import os

from adafri.v1.base.firebase_collection import FirebaseCollectionBase

# Set your JWKS URI and issuer (OIDC provider)
CLIENT_ID = os.environ.get('ADAFRI_CLIENT_ID', None)
AUDIENCE = os.environ.get('OAUTH_AUDIENCE', CLIENT_ID)

def getDefaultClaims(well_known, audience=None):
    default_claims = {
        "iss": {"essential": True, "value": well_known['issuer']},
        "aud": {"essential": True, "value": audience},
    }
    return default_claims
def fetch_well_known(well_known_uri):
    try:
        response = requests.get(well_known_uri)
        response.raise_for_status()
        return response.json()
    except Exception as e:
        print('exception occurated while fetching well_known uri')
        return None
def fetch_jwks(jwks_uri):
    try:
        response = requests.get(jwks_uri)
        response.raise_for_status()
        return response.json()
    except Exception as e:
        print('exception occurated while fetching jwks_uri')
        return None

# Step 2: Verify the JWT token
def verify_jwt_token(token, jwks, issuer, audience, algorithms=['RS256'], claims_options=None):
    # Create a JsonWebToken instance for decoding and verification
    jwt = JsonWebToken(algorithms=algorithms)  # Specify the algorithm(s) used by your provider
    key = KeySet(jwks)
    # Verify the token using the JWKS, issuer, and audience
    claims = jwt.decode(
        token,
        key.keys,
        claims_options=claims_options,
    )
    # print('algs', algorithms)
    # Ensure the token is not expired and other claims are valid
    claims.validate()
    # print('keys', claims)
    return claims

# Fetch JWKS and verify the token
from datetime import datetime

def check_token(token, well_known_uri=None, audience=AUDIENCE):
    if well_known_uri is None:
        well_known_uri = os.environ.get('WELL_KNOWN_URI', None)
    if well_known_uri is None:
        return None
    well_known = fetch_well_known(well_known_uri)
    if well_known is None:
        return None
    if 'issuer' in well_known and 'jwks_uri' in well_known and 'id_token_signing_alg_values_supported' in well_known:
        try:
            # jwks = fetch_jwks(jwks_uri)
            jwks = fetch_jwks(well_known['jwks_uri'])
            # print('jwks', jwks)
            algorithms = well_known['id_token_signing_alg_values_supported']
            claims = verify_jwt_token(token, jwks, well_known['issuer'], audience, algorithms, getDefaultClaims(well_known, audience=audience))
            # print('claims', claims)
            if claims is not None and 'iss' in claims:
                exp = datetime.fromtimestamp(claims['exp'])
                iat = datetime.fromtimestamp(claims['iat'])
                print('exp', exp)
                print('iat', iat)
                return claims
        except Exception as e:
            print("Token verification failed:", e)
            return None
    return None

def is_existing_token(_token: str):
    db = FirebaseCollectionBase(collection_name='clients_token_collection')
    token = db.custom_query([{"key": "access_token", "comp": "==", "value": _token}], True)
    if token is not None:
        return token
    return None


from glom import glom
from typing import List, Dict, Any, Union
from glom import glom

def convert_paths_to_glom_spec(paths: List[str]) -> Dict:
    """
    Convert dot notation paths to glom specification.
    
    Args:
        paths: List of dot notation paths (e.g., ['users.name', 'users.profile.age'])
        
    Returns:
        Dict: Glom specification
    """
    def is_array_path(path_parts: List[str], all_paths: List[List[str]]) -> bool:
        """
        Determine if a path is an array path by checking if any other paths
        extend from this path.
        """
        current_path = '.'.join(path_parts)
        for other_parts in all_paths:
            other_path = '.'.join(other_parts)
            if other_path != current_path and other_path.startswith(current_path + '.'):
                return True
        return False

    def group_paths_by_root(paths: List[str]) -> Dict[str, List[List[str]]]:
        """Group paths by their root key."""
        grouped = {}
        for path in paths:
            parts = path.split('.')
            root = parts[0]
            if root not in grouped:
                grouped[root] = []
            grouped[root].append(parts)
        return grouped

    def build_nested_spec(paths: List[List[str]], all_paths: List[List[str]]) -> Dict:
        """Build nested spec structure from paths."""
        spec = {}
        
        for parts in paths:
            current = spec
            path_length = len(parts)
            
            # Build the path incrementally
            for i, part in enumerate(parts[1:], 1):
                subpath = parts[:i+1]
                
                if i == path_length - 1:  # Last part
                    if is_array_path(parts[:-1], all_paths):
                        # Parent is an array
                        parent_path = '.'.join(parts[1:-1])
                        if parts[-2] not in current:
                            current[parts[-2]] = (parent_path, [{part: part}])
                    else:
                        # Regular field
                        current[part] = '.'.join(parts[1:])
                else:
                    if is_array_path(subpath, all_paths):
                        # Current part is an array
                        if part not in current:
                            current[part] = ('.'.join(parts[1:i+1]), [{}])
                        current = current[part][1][0]
                    else:
                        # Regular nested object
                        if part not in current:
                            current[part] = {}
                        current = current[part]
        
        return spec

    # Get all path parts for analysis
    all_path_parts = [path.split('.') for path in paths]
    
    # Group paths by root
    grouped_paths = group_paths_by_root(paths)
    
    # Build the final spec
    result = {}
    for root, paths_list in grouped_paths.items():
        nested_spec = build_nested_spec(paths_list, all_path_parts)
        result[root] = (root, [nested_spec])
        
    return result
data = {
    'users': [
        {
            'id': 1,
            'name': 'John',
            'profile': {
                'age': 30,
                'contacts': [
                    {'type': 'email', 'value': 'john@example.com'},
                    {'type': 'phone', 'value': '123-456-7890'}
                ]
            }
        },
        {
            'id': 2,
            'name': 'Jane',
            'profile': {
                'age': 25,
                'contacts': [
                    {'type': 'email', 'value': 'jane@example.com'},
                    {'type': 'phone', 'value': '098-765-4321'}
                ]
            }
        }
    ]
}

def pick(obj: Any, paths: Union[List[str], str]) -> Any:
    """
    Pick specific fields from nested dictionaries/lists using dot notation.
    Only shows specified fields in nested arrays and objects.
    
    Args:
        obj: Source object (dict or list)
        paths: Single path string or list of paths using dot notation
               e.g., 'users.name' or ['users.name', 'users.profile.age']
    
    Returns:
        Dict/List with only the specified fields
    """
    # Handle single path string
    if isinstance(paths, str):
        paths = [paths]
        
    # Handle lists
    if isinstance(obj, list):
        return [pick(item, paths) for item in obj]
        
    # Handle non-dict values
    if not isinstance(obj, dict):
        return obj
        
    # Group paths by their first part and track array paths
    grouped_paths = {}
    array_paths = set()
    
    for path in paths:
        parts = path.split('.')
        if not parts:
            continue
            
        root = parts[0]
        if root not in grouped_paths:
            grouped_paths[root] = []
            
        if len(parts) > 1:
            subpath = '.'.join(parts[1:])
            grouped_paths[root].append(subpath)
            
            # Track array paths
            if isinstance(obj.get(root), list):
                array_paths.add(root)
            elif len(parts) > 2 and isinstance(obj.get(parts[0], {}).get(parts[1]), list):
                array_paths.add(f"{parts[0]}.{parts[1]}")
            
    # Build result
    result = {}
    for key, subpaths in grouped_paths.items():
        value = obj.get(key)
        if value is not None:
            if subpaths:
                picked = pick(value, subpaths)
                # Only include non-empty results
                if picked not in (None, {}, []):
                    result[key] = picked
            else:
                result[key] = value
                
    return result
r = pick(data, ['users.name', 'users.profile.contacts'])
print(r)

# # Pick names only
# result1 = glom(data, convert_paths_to_glom_spec(['users.name', 'users.profile.contacts.type']))
# print("Names only:", result1)



from typing import Any
from dataclasses import dataclass
import json
from .utils import JsonEncoder
from enum import Enum

class ResponseStatus:
    OK = 'ok'
    ERROR = 'error'

class StatusCode:
    status_200 = 200
    status_400 = 400

@dataclass
class Error:
    message: str
    name: str
    error_code: int = 1

    @staticmethod
    def from_dict(obj: Any) -> 'Error':
        _message = str(obj.get("message"))
        _name = str(obj.get("name"))
        _error_code = "1"
        code = obj.get("error_code")
        if code is not None:
            _error_code = int(str(code))
        return Error(_message, _name, _error_code)

@dataclass
class ApiResponse:
    status: str
    status_code: int
    data: Any
    error: Error

    @staticmethod
    def from_dict(obj: Any) -> 'ApiResponse':
        _status = str(obj.get("status"))
        _status_code = int(obj.get("status_code"))
        _data = obj.get("data")
        _error = Error.from_dict(obj.get("error"))
        return ApiResponse(_status, _status_code, _data, _error)
    
    def to_json(self):
        return self.to_dict()

    def to_dict(self, fields=None):
        values = {
            "status": self.status,
            "status_code": self.status_code,
            "data": self.data,
            "error": self.error
        }
        if 'data' in values and values['data'] is None:
            del values['data']
        if 'error' in values and values['error'] is None:
            del values['error']
        return values

    # def to_json(self):
    #     return json.loads(JsonEncoder().encode(self))



from .objectives import objectives_data
from .demographics import genders_data, ages_data
from .adschedules import times_data, days_data
from .bids import bids_data
from .adsChannel import ads_channel_data
from .devices import devices_data


from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass, BaseClass,init_class_kwargs,get_object_model_class
        
@dataclass
class AssetTextFields(BaseFieldsClass):
    assetText = "assetText"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
@dataclass(init=False)
class AssetText(BaseClass):
    assetText: str
    def __init__(self, asset=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, asset, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "assetText": self.assetText,
        })
    @staticmethod
    def from_dict(obj: Any) -> 'AssetText':
        cls_object, keys = get_object_model_class(obj, AssetText, None);
        return AssetText(cls_object)





from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass, BaseClass, init_class_kwargs,get_object_model_class
@dataclass
class AgesFields(BaseFieldsClass):
    text = "text"
    ageRangeType = "ageRangeType"
    criterionId = "criterionId"
    criterionType = "criterionType"
    type = "type"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

@dataclass(init=False)
class Ages(BaseClass):
    text: str
    ageRangeType: str
    criterionId: int
    criterionType: str
    type: str
    def __init__(self, ages=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, ages, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "text": self.text,
            "ageRangeType": self.ageRangeType,
            "criterionId": self.criterionId,
            "criterionType": self.criterionType,
            "type": self.type
        })
    @staticmethod
    def from_dict(obj: Any) -> 'Ages':
        cls_object, keys = get_object_model_class(obj, Ages, None);
        return Ages(cls_object)


from .....utils import (DictUtils, get_object_model_class, init_class_kwargs, BaseClass, get_request_fields, RequestFields)
from .....utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from typing import Any
from dataclasses import dataclass
import pydash
from ...models.adgroups.base import BaseAdGroup
from ...models.placement import (Website, YoutubeChannel, YoutubeVideo)
from .adgroup_fields import (AdGroupDisplayFields, AdGroupDisplayFieldsProps, ADGROUP_DISPLAY_PICKABLE_FIELDS, DISPLAY_ADGROUPS_COLLECTION)
from ...models.placement import Website, YoutubeChannel, YoutubeVideo
from ...models.display_ads import (DisplayAdsToPublish, ResponsiveDisplayAdsToPublish)

def filter_adgroup_display_request_fields(fields, default_fields = ADGROUP_DISPLAY_PICKABLE_FIELDS):
    request_fields = get_request_fields(fields, default_fields, [default_fields[0]])
    if request_fields is None or bool(request_fields) is False:
        request_fields = [default_fields[0]]
    return request_fields

@dataclass(init=False)
class AdGroupDisplay(BaseAdGroup):
    targetedPlacements: list[Website]
    excludedPlacements: list[Website]
    websites: list[Website]
    youtubeChannels: list[YoutubeChannel]
    youtubeVideos: list[YoutubeVideo]
    __baseFields: AdGroupDisplayFieldsProps

    def __init__(self, adgroup=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, adgroup, ADGROUP_DISPLAY_PICKABLE_FIELDS, AdGroupDisplayFieldsProps, DISPLAY_ADGROUPS_COLLECTION, ['id'], **kwargs)
        super().__init__(**{**data_args, "adgroup": adgroup});
        for key in keys:
            if AdGroupDisplayFieldsProps.get(key) is None:
                website = ['targetedPlacements', 'excludedPlacements', 'websites']
                if key in website:
                    setattr(self, key, Website().fromListDict(cls_object[key]))
                elif key == 'youtubeChannels':
                    setattr(self, key, YoutubeChannel().fromListDict(cls_object[key]))
                elif key == 'youtubeVideos':    
                    setattr(self, key, YoutubeVideo().fromListDict(cls_object[key]))
                elif key == 'images':
                    setattr(self, key, DisplayAdsToPublish().fromListDict(cls_object[key]))
                elif key == 'imagesNative':
                    setattr(self, key, ResponsiveDisplayAdsToPublish().fromListDict(cls_object[key]))
                else:
                    setattr(self, key, cls_object[key]) 


    @staticmethod
    def generate_model(_key_="default_value"):
        user = {};
        props = AdGroupDisplayFieldsProps
        for k in DictUtils.get_keys(props):
            user[k] = props[k][_key_];
        return user;

    @staticmethod
    def from_dict(adgroup: Any, db=None, collection_name=None) -> 'AdGroupDisplay':
        cls_object, keys = get_object_model_class(adgroup, AdGroupDisplay, AdGroupDisplayFieldsProps);
        _adgroup = AdGroupDisplay(cls_object, db=db, collection_name=collection_name)
        return _adgroup
    
    def to_dict(self, fields=None):
        base = super().to_dict(fields)
        return {
            **base,
            "targetedPlacements": Website().toListDict(self.targetedPlacements, None),
            "excludedPlacements": Website().toListDict(self.excludedPlacements, None),
            "websites": Website().toListDict(self.websites, None),
            "youtubeChannels": YoutubeChannel().toListDict(self.youtubeChannels, None),
            "youtubeVideos": YoutubeVideo().toListDict(self.youtubeVideos, None),
            "images": DisplayAdsToPublish().toListDict(getattr(self, 'images', None), None),
            "imagesNative": ResponsiveDisplayAdsToPublish().toListDict(getattr(self, 'imagesNative', None), None),
    }
    def getAdGroups(self, accountId, campaignId, fields=None):
        print('getAdGroups', accountId, campaignId)
        isID = False
        try:
            campaignId = int(campaignId)
        except:
            isID = True
        if isID is False and accountId is None:
            return None
        adgroups = []
        adgroups_ = self.query(AdGroupDisplay, "adgroup", query_params=[{"key": "accountId", "comp": "==", "value": accountId}, {"key": "campaign_id", "comp": "==", "value": campaignId}])
        request_fields = filter_adgroup_display_request_fields(fields)
        print('request_fields adgroup ===>', request_fields)
        # print('account query', account)
        for adgroup in adgroups_:
            adgroups.append(DictUtils.filter_by_fields(adgroup.to_dict(), request_fields))

        return adgroups
    
    def get(self, fields=None):
        if bool(self.id) is None:
            return None;
        request_fields = filter_adgroup_display_request_fields(fields)
        doc = self.document_reference().get();
        if doc.exists is False:
            return None;
        data = {"id": doc.id, **doc.to_dict()}
        return AdGroupDisplay.from_dict(adgroup=AdGroupDisplay.from_dict(adgroup=data).to_dict(),  db=self.db, collection_name=self.collection_name);

    def getByUserId(self,id, fields=None):
        adgroups = [];
        request_fields = filter_adgroup_display_request_fields(fields)
        docs = self.collection().where('owner','==',id).get();
        for doc in docs:
            data = doc.to_dict();
            data['id'] = doc.id;
            adgroups.append(AdGroupDisplay.from_dict(adgroup=AdGroupDisplay.from_dict(adgroup=data).to_json(request_fields),  db=self.db, collection_name=self.collection_name));
        return adgroups;


    def update(self, data):
        try:
            last_value = self.to_json();
            filtered_value = pydash.pick(data, AdGroupDisplayFields.filtered_keys('editable', True));
            new_value = DictUtils.merge_dict(filtered_value, self.to_json());
            changed_fields = DictUtils.get_changed_field(last_value, new_value);
            data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
            if bool(data_update) is False:
                return None;
            self.document_reference().set(data_update, merge=True)
            return DictUtils.dict_from_keys(self.getAccount(), changed_fields);
        except Exception as e:
            print(e)
            return None;

    def remove(self, only_mark_as_removed=True):
        try:
            if self.id is None:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot identify Campaign with id {self.id}","INVALID_REQUEST", 1));
            if only_mark_as_removed:
                self.document_reference().set({"is_removed": True})
            else:
                self.document_reference().delete();
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"Campaign {self.id} deleted"}, None);
        except:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated while removing authorization code with id {self.id}","INVALID_REQUEST", 1));




from .....utils import DictUtils, Crypto, get_object_model_class, pydash, init_class_kwargs
from ....base.firebase_collection import (FirebaseCollectionBase, getTimestamp)
from .token_fields import TokenFields, TokenFieldsProps, STANDARD_FIELDS, TOKEN_COLLECTION, get_token_expire_at, is_expired
from .....utils.response import ApiResponse, Error, ResponseStatus, StatusCode
from .....utils.utils import read_file
from typing import Any
from dataclasses import dataclass
from ....user import User
from authlib.oauth2.rfc7009 import RevocationEndpoint
from authlib.oauth2.rfc6750 import BearerTokenValidator
from authlib.oauth2.rfc6749 import grants
from authlib.oauth2.rfc6749 import TokenMixin, scope_to_list
from authlib.oauth2.rfc6750 import BearerTokenGenerator, BearerToken
from authlib.oauth2.rfc9068 import JWTBearerTokenGenerator, JWTBearerTokenValidator
import json
from flask import abort, Response
from joserfc.jwk import KeySet

@dataclass(init=False)
class OAuthToken(TokenMixin, FirebaseCollectionBase):
    id: str
    client_id: str
    uid: str
    token_type: str
    access_token: str
    refresh_token: str
    scopes: list[str]
    scope: str
    expired_at: str
    expires_in: int
    revoked: bool
    type: str
    __baseFields: TokenFieldsProps

    def __init__(self, token=None, **kwargs):
        if type(token) is str:
            token = {"access_token": token} 
        (cls_object, keys, data_args) = init_class_kwargs(self, token, STANDARD_FIELDS, TokenFieldsProps, TOKEN_COLLECTION, ['id'], **kwargs)
        super().__init__(**data_args);
        for key in keys:
            setattr(self, key, cls_object[key]) 


    @staticmethod
    def generate_model(_key_="default_value"):
        grant = {};
        props = TokenFieldsProps
        for k in DictUtils.get_keys(props):
            grant[k] = props[k][_key_];
        return grant;

    @staticmethod
    def from_dict(token: Any=None, db=None, collection_name=None) -> 'OAuthToken':
        cls_object, keys = get_object_model_class(token, OAuthToken, TokenFieldsProps);
        _client = OAuthToken(cls_object, db=db, collection_name=collection_name)
        return _client

    def query(self, query_params: list, first=False, limit=None):
        result = [];
        query_result = self.custom_query(query_params, first=first, limit=limit)
        if bool(query_result):
            if first is True:
                return OAuthToken.from_dict(token=query_result, db=self.db, collection_name=self.collection_name)
            else:
                for doc in query_result:
                    result.append(OAuthToken.from_dict(token=doc, db=self.db, collection_name=self.collection_name))
                return result
        if first:
                return None
        return [];

    def getOAuthToken(self) -> 'OAuthToken':
        if bool(self.id):
            doc = self.document_reference().get();
            if doc.exists is False:
                return None;
            return OAuthToken.from_dict(doc.to_dict(), db=self.db, collection_name=self.collection_name);
        if bool(self.access_token):
            return self.query([{"key": TokenFields.access_token, "comp": "==", "value": self.access_token}], True)

    @staticmethod
    def generate(**kwargs) -> 'ApiResponse':
        data_dict = DictUtils.pick_fields(kwargs, TokenFields.filtered_keys('pickable', True));
        token_model = OAuthToken.from_dict(DictUtils.merge_dict(data_dict, OAuthToken.generate_model()));
        
        if bool(token_model.to_json()) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Empty request","INVALID_REQUEST", 1)).to_json()
        
        if bool(token_model.access_token) is False:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("name required","INVALID_REQUEST", 1));

        token_model.id = Crypto().generate_id(token_model.uid+"~"+token_model.client_id+"~"+token_model.type);
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, token_model, None);
    

    def save(self, token, request):
        model = {**token, "client_id": request.client.client_id, "uid": request.user.uid, "revoked": False}
        if 'type' not in model:
            model['type'] = 'app_token'
        print('saving token', model)
        token_generate = OAuthToken.generate(**model);
        if token_generate.status == ResponseStatus.ERROR:
            return token_generate
        token_model: OAuthToken = token_generate.data;
        docRef = OAuthToken(token_model.to_json()).document_reference();
        # if docRef.get().exists:
        #     return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Location with name {token_generate.data.id} already exist","INVALID_REQUEST", 1));
        
        docRef.set({**token_model.to_json(), "createdAt": getTimestamp()}, merge=True);
        created_token = token_model.getOAuthToken()
        return ApiResponse(ResponseStatus.OK, StatusCode.status_200, created_token.to_json(), None);
    
    def update(self, data):
        try:
            last_value = self.to_json();
            filtered_value = pydash.pick(data, TokenFields.filtered_keys('editable', True));
            new_value = DictUtils.merge_dict(filtered_value, self.to_json());
            changed_fields = DictUtils.get_changed_field(last_value, new_value);
            data_update = DictUtils.dict_from_keys(filtered_value, changed_fields);
            if bool(data_update) is False:
                return None;
            self.document_reference().set(data_update, merge=True)
            return DictUtils.dict_from_keys(self.getOAuthToken().to_json(), changed_fields);
        except Exception as e:
            print(e)
            return None;
    
    def remove(self):
        try:
            if self.id is None:
                return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"Cannot identify token with id {self.id} to delete","INVALID_REQUEST", 1));
            deleted = self.document_reference().delete();
            return ApiResponse(ResponseStatus.OK, StatusCode.status_200, {"message": f"Token {self.id} deleted"}, None);
        except:
            return ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error(f"An error occurated while removing authorization code with id {self.id}","INVALID_REQUEST", 1));


    def is_expired(self):
        return is_expired(self.expired_at)
    
    def is_revoked(self):
        return self.revoked
    

class TokenValidator(BearerTokenValidator):
    def authenticate_token(self, token_string):
        token_request = OAuthToken().query([{"key":"access_token", "comp": "==", "value": token_string}], True)
        return token_request
    
    def validate_token(self, token, scopes, request):
        token_scopes = scope_to_list(token.scope);
        insufficient = self.scope_insufficient(token_scopes, scopes);
        if insufficient:
            response = ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Insufficient privilegies","INVALID_REQUEST", 1)).to_json()
            return abort(Response(response=json.dumps(response), status=401, mimetype='application/json'))        
        return None

class TokenRevocationEndpoint(RevocationEndpoint):
    def query_token(self, token, token_type_hint, client):
        q: list = OAuthToken().query([{"key":"client_id", "comp": "==", "value": client.clent_id}], False)
        if token_type_hint == 'access_token':
            return pydash.find(q, lambda x: x.access_token==token);
        elif token_type_hint == 'refresh_token':
            return pydash.find(q, lambda x: x.refresh_token==token);
        # without token_type_hint
        item = pydash.find(q, lambda x: x.access_token==token);
        if item:
            return item
        return pydash.find(q, lambda x: x.refresh_token==token);

    def revoke_token(self, _token):
        token = OAuthToken.from_dict(_token);
        token.revoked = True
        token.update(_token);

class RefreshTokenGrant(grants.RefreshTokenGrant):
    def authenticate_refresh_token(self, refresh_token):
        token = OAuthToken().query([{"key":"refresh_token", "comp": "==", "value": refresh_token}], True);
        return token;
        # if token and token.is_refresh_token_active():
        #     return token

    def authenticate_user(self, credential):
        return User({"uid": credential.user_id}).get()

    # def revoke_old_credential(self, credential):
    #     credential.revoked = True
    #     db.session.add(credential)
    #     db.session.commit()


DEFAULT_EXPIRES_IN = 3600
class TokenGenerator(BearerTokenGenerator):
    @staticmethod
    def generate(grant_type, client, user=None, scope=None, expires_in=None, include_refresh_token=True):
        if expires_in is None:
            expires_in = DEFAULT_EXPIRES_IN
        uid = client.uid;
        if user is not None:
            uid = user.uid
        expires_at = get_token_expire_at(expires_in).isoformat()
        token = {'token_type': 'Bearer', "client_id": client.client_id, "uid": uid, 'scope': scope, 'scopes': scope_to_list(scope), 'expires_in': expires_in, "grant_type": grant_type, 'expired_at': expires_at}
        access_token = Crypto().generate_token("access_token~"+json.dumps(token));
        token['access_token'] = access_token;
        if include_refresh_token:
            token['refresh_token'] = Crypto().generate_token("refresh_token~"+json.dumps(token));
        if grant_type == 'implicit':
            return {"access_token": access_token, "scope": scope, "expires_in": expires_in}
        print('token generated', access_token)
        return token
import os
import json
from authlib.jose import jwt
import time
from datetime import datetime, timedelta
from joserfc.jwk import RSAKey
import requests


def read_private_key():
    pem_url = os.environ.get('OIDC_PEM_URL', None)
    json_url = os.environ.get('OIDC_JSON_URL', None)
    local_pem_file = os.environ.get('LOCAL_OIDC_PEM_URL', None)
    local_json_file = os.environ.get('LOCAL_OIDC_JSON_URL', None)
    # print('private key search', {"local_pem": local_pem_file, "pem_url": pem_url, "local_json": local_json_file, "json_url": json_url})
    if local_pem_file is not None:
        try:
            fp = os.getcwd() + "/" + local_pem_file
            # print('file for local pem', fp)
            key = import_key(fp, False)
            if key is None:
                print('import returns none')
            if key is not None:
                print('import returns success')
                return key
            else:
                print(f'cannot load private pem key from local file {local_pem_file}')
        except Exception as e:
            print(f'exception occurated while loading private pem key from local file {local_pem_file} ', e)
            pass
    if local_json_file is not None:
        try:
            fp = os.getcwd() + "/" + local_json_file
            # print('file for local json', fp)
            key = import_key(fp, True)
            if key is not None:
                return key
            else:
                print(f'cannot load private json key from local file {local_json_file} ')
        except Exception as e:
            print(f'exception occurated while loading private json key from local file {local_json_file} ', e)
            pass

    if pem_url is not None and len(pem_url) > 0 and json_url is None:
        response = requests.get(pem_url)
        # print('pem request response status code ', response.status_code)
        if response.status_code == 200:
            val = json.loads(response.content)
            # print('val', val)
            # print('type val', type(val))
            return val
    if json_url is not None and len(json_url) > 0:
        response = requests.get(json_url)
        # print('json request response status code ', response.status_code)
        if response.status_code == 200:
            return json.loads(response.content)
    return None

def read_jwks():
    """
    Reads JWKS from either a local file or a URL.

    First it looks for a local file specified in the environment variable
    LOCAL_OIDC_JWKS_URL. If not found, it looks for a URL specified in the
    environment variable OIDC_JWKS_URL. If neither is found, it returns None.

    If a local file is found, it reads and parses the content as JSON.
    If a URL is found, it sends a GET request to the URL and parses the
    response content as JSON.

    If an exception occurs while reading the file or sending the request,
    it prints an error message and returns None.

    :return: JWKS as a JSON object
    :rtype: dict or None
    """
    url = os.environ.get('OIDC_JWKS_URL', None)
    local_jwks_file = os.environ.get('LOCAL_OIDC_JWKS_URL', None)
    if local_jwks_file is not None:
        try:
            fp = os.getcwd() + "/" + local_jwks_file
            content = open(fp, 'r').read()
            return json.loads(content)
        except Exception as e:
            print(f'exception occurated while loading jwk from local file {local_jwks_file} ', e)
            return None
    if url is not None and len(url) > 0:
        response = requests.get(url)
        # print('jwks fetch response code', response.status_code)
        if response.status_code == 200:
            return response.json()


class JwtTokenGenerator(JWTBearerTokenGenerator):
    GRANT_TYPES_EXPIRES_IN = {
        "authorization_code": 864000,
        "implicit": 3600,
        "password": 864000,
        "client_credentials": 864000,
    }
    def get_jwks(self):
        print('reading jwks for generator [JwtTokenGenerator]')
        return read_private_key()


    # def access_token_generator(self, client, grant_type, user, scope):
    #     # now = datetime.now()
    #     # exp = now + timedelta(seconds=3600)
    #     # expires_in = now + self._get_expires_in(client, grant_type)
    #     now = int(time.time())
    #     exp = self._get_expires_in(client, grant_type)
    #     expires_in = now + self._get_expires_in(client, grant_type)
    #     print('now', now)
    #     print('exp', exp)
    #     print('expires_in', expires_in)

    #     token_data = {
    #         'iss': self.issuer,
    #         'exp': expires_in,
    #         'client_id': client.get_client_id(),
    #         'iat': now,
    #         'jti': self.get_jti(client, grant_type, user, scope),
    #         'scope': scope,
    #     }

    #     # In cases of access tokens obtained through grants where a resource owner is
    #     # involved, such as the authorization code grant, the value of 'sub' SHOULD
    #     # correspond to the subject identifier of the resource owner.

    #     if user:
    #         token_data['sub'] = user.get_user_id()

    #     # In cases of access tokens obtained through grants where no resource owner is
    #     # involved, such as the client credentials grant, the value of 'sub' SHOULD
    #     # correspond to an identifier the authorization server uses to indicate the
    #     # client application.

    #     else:
    #         token_data['sub'] = client.get_client_id()

    #     # If the request includes a 'resource' parameter (as defined in [RFC8707]), the
    #     # resulting JWT access token 'aud' claim SHOULD have the same value as the
    #     # 'resource' parameter in the request.

    #     # TODO: Implement this with RFC8707
    #     if False:  # pragma: no cover
    #         ...

    #     # If the request does not include a 'resource' parameter, the authorization
    #     # server MUST use a default resource indicator in the 'aud' claim. If a 'scope'
    #     # parameter is present in the request, the authorization server SHOULD use it to
    #     # infer the value of the default resource indicator to be used in the 'aud'
    #     # claim. The mechanism through which scopes are associated with default resource
    #     # indicator values is outside the scope of this specification.

    #     else:
    #         token_data['aud'] = self.get_audiences(client, user, scope)

    #     # If the values in the 'scope' parameter refer to different default resource
    #     # indicator values, the authorization server SHOULD reject the request with
    #     # 'invalid_scope' as described in Section 4.1.2.1 of [RFC6749].
    #     # TODO: Implement this with RFC8707

    #     if auth_time := self.get_auth_time(user):
    #         token_data['auth_time'] = auth_time

    #     # The meaning and processing of acr Claim Values is out of scope for this
    #     # specification.

    #     if acr := self.get_acr(user):
    #         token_data['acr'] = acr

    #     # The definition of particular values to be used in the amr Claim is beyond the
    #     # scope of this specification.

    #     if amr := self.get_amr(user):
    #         token_data['amr'] = amr

    #     # Authorization servers MAY return arbitrary attributes not defined in any
    #     # existing specification, as long as the corresponding claim names are collision
    #     # resistant or the access tokens are meant to be used only within a private
    #     # subsystem. Please refer to Sections 4.2 and 4.3 of [RFC7519] for details.

    #     token_data.update(self.get_extra_claims(client, grant_type, user, scope))

    #     # This specification registers the 'application/at+jwt' media type, which can
    #     # be used to indicate that the content is a JWT access token. JWT access tokens
    #     # MUST include this media type in the 'typ' header parameter to explicitly
    #     # declare that the JWT represents an access token complying with this profile.
    #     # Per the definition of 'typ' in Section 4.1.9 of [RFC7515], it is RECOMMENDED
    #     # that the 'application/' prefix be omitted. Therefore, the 'typ' value used
    #     # SHOULD be 'at+jwt'.

    #     header = {'alg': self.alg, 'typ': 'at+jwt'}
    #     jwks = self.get_jwks()
    #     access_token = jwt.encode(
    #         header,
    #         token_data,
    #         key=jwks,
    #         check=False,
    #     )
    #     return access_token.decode()
    # def generate_token(self, client, grant_type, user=None, scope=None):
    #     """Generate a JWT as a bearer token."""
    #     now = int(time.time())
    #     header = {"alg": self.algorithm}
    #     print('generate token grant type', grant_type)
        
    #     # Define payload for the JWT
    #     payload = {
    #         'iss': self.issuer,
    #         'aud': self.audience,
    #         'iat': now,
    #         'exp': now + 3600,  # Token expiration (1 hour)
    #         'client_id': client.client_id,
    #         'scope': scope,
    #     }
        
    #     if user:
    #         payload['uid'] = user.uid
        
    #     # Create the JWT
    #     json_url = os.environ.get('OIDC_JSON_URL', None)
    #     response = requests.get(json_url)
    #     # rsa = RSAKey.import_key(open(os.getcwd()+"/"+os.getenv('OAUTH_PRIVATE_KEY'), 'r').read())
    #     rsa = RSAKey.import_key(response.json(), **{"use": "sig"})
    #     private_jwk = rsa.as_dict(is_private=True)
    #     token = jwt.encode(header, payload, private_jwk)
    #     return token.decode('utf-8')  # Return token as string
    # # @staticmethod
    # def generate(self, grant_type, client, user=None, scope=None,
    #         expires_in=None, include_refresh_token=True):
    #     """Generate a bearer token for OAuth 2.0 authorization token endpoint.

    #     :param client: the client that making the request.
    #     :param grant_type: current requested grant_type.
    #     :param user: current authorized user.
    #     :param expires_in: if provided, use this value as expires_in.
    #     :param scope: current requested scope.
    #     :param include_refresh_token: should refresh_token be included.
    #     :return: Token dict
    #     """
    #     print('generating new token')
    #     scope = self.get_allowed_scope(client, scope)
    #     access_token = self.access_token_generator(
    #         client=client, grant_type=grant_type, user=user, scope=scope)
    #     if expires_in is None:
    #         expires_in = self._get_expires_in(client, grant_type)

    #     token = {
    #         'token_type': 'Bearer',
    #         'access_token': access_token,
    #     }
    #     if expires_in:
    #         token['expires_in'] = expires_in
    #     if include_refresh_token and self.refresh_token_generator:
    #         token['refresh_token'] = self.refresh_token_generator(
    #             client=client, grant_type=grant_type, user=user, scope=scope)
    #     if scope:
    #         token['scope'] = scope
    #     return token

class JwtTokenValidator(JWTBearerTokenValidator):
    def get_jwks(self):
        # print('reding jwks for validator [JwtTokenValidator]')
        return read_private_key()
        # jwks = read_jwks()
        # for k in jwks['keys']:
        #     if k['kty'] == 'RSA':
        #         print('find rsa for validator', k)
        #         return RSAKey.import_key(k).as_dict(private=False)
        # return None

    def authenticate_token(self, token_string):
        token_request = OAuthToken().query([{"key":"access_token", "comp": "==", "value": token_string}], True)
        return token_request
    
    def validate_token(self, token, scopes, request):
        if token is None:
            response = ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Insufficient privilegies","ACCESS_DENIED", 1)).to_json()
            return abort(Response(response=json.dumps(response), status=401, mimetype='application/json')) 
        token_scopes = scope_to_list(token.scope);
        insufficient = self.scope_insufficient(token_scopes, scopes);
        if insufficient:
            response = ApiResponse(ResponseStatus.ERROR, StatusCode.status_400, None, Error("Insufficient privilegies","ACCESS_DENIED", 1)).to_json()
            return abort(Response(response=json.dumps(response), status=401, mimetype='application/json'))        
        return None
# Define your JWT bearer token generator
# class JWTBearerTokenGenerator(BearerToken):
#     def __init__(self, private_key, algorithm='RS256', issuer=None, audience=None):
#         self.private_key = private_key
#         self.algorithm = algorithm
#         self.issuer = issuer
#         self.audience = audience

#     def generate(self, grant_type, client, user=None, scope=None, expires_in=None, include_refresh_token=True):
#         v = self.generate()
#     def generate_token(self, client, grant_type, user=None, scope=None):
#         """Generate a JWT as a bearer token."""
#         now = int(time.time())
#         header = {"alg": self.algorithm}
        
#         # Define payload for the JWT
#         payload = {
#             'iss': self.issuer,
#             'aud': self.audience,
#             'iat': now,
#             'exp': now + 3600,  # Token expiration (1 hour)
#             'client_id': client.client_id,
#             'scope': scope,
#         }
        
#         if user:
#             payload['uid'] = user.uid
        
#         # Create the JWT
#         token = jwt.encode(header, payload, self.private_key)
#         return token.decode('utf-8')  # Return token as string

#     def get_token(self, client, grant_type, user=None, scope=None):
#         """Generate a token and set token type as Bearer."""
#         access_token = self.generate_token(client, grant_type, user, scope)
#         return {
#             'access_token': access_token,
#             'token_type': 'Bearer',
#             'expires_in': 3600,  # Token validity duration
#             'scope': scope
#         }


devices_data = [
  {
    "id": 30000,
    "criterionId": 30000,
    "name": 'Ordinateur',
    "type": 'PLATFORM',
    "criterionType": 'PLATFORM',
    "isTargeted": True,
    "isExcluded": False
  },
  {
    "id": 30001,
    "criterionId": 30001,
    "name": 'Smartphone',
    "type": 'PLATFORM',
    "criterionType": 'PLATFORM',
    "isTargeted": True,
    "isExcluded": False
  },
  {
    "id": 30002,
    "criterionId": 30002,
    "name": 'Tablette',
    "type": 'PLATFORM',
    "criterionType": 'PLATFORM',
    "isTargeted": True,
    "isExcluded": False
  },
  {
    "id": 30004,
    "criterionId": 30004,
    "name": 'Smart Tv',
    "type": 'PLATFORM',
    "criterionType": 'PLATFORM',
    "isTargeted": True,
    "isExcluded": False
  }
]


from dataclasses import dataclass
from typing import Any
from  ....utils.utils import DictUtils, BaseFieldsClass, BaseClass,init_class_kwargs,get_object_model_class

@dataclass
class DisplayAdsToPublishFields(BaseFieldsClass):
    id = "id"
    name = "name"
    mediaId = "mediaId"
    type = "type"
    width = "width"
    height = "height"
    url = "url"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

@dataclass
class AssetNativeFields(BaseFieldsClass):
    id = "id"
    assetId = "assetId"
    width = "width"
    height = "height"
    imageUrl = "imageUrl"
    imageFileSize = "imageFileSize"
    imageMimeType = "imageMimeType"
    usage = "usage"
    useFor = "useFor"
    owner = "owner"
    aacid = "aacid"
    clientCustomerId = "clientCustomerId"
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@dataclass(init=False)
class AssetNative(BaseClass):
    id: str
    name: str
    assetId: int
    imageFileSize: int
    imageMimeType: str
    usage: list[str]
    useFor: str
    owner: str
    aacid: str
    width: int
    height: int
    imageUrl: str
    def __init__(self, asset=None, **kwargs):
        (cls_object, keys, data_args) = init_class_kwargs(self, asset, None, None, None, [], **kwargs)
        for key in keys:
            setattr(self, key, cls_object[key]) 
    def to_dict(self, field=None):
        return DictUtils.remove_none_values({
            "id": self.id,
            "name": self.name,
            "assetId": self.assetId,
            "imageFileSize": self.imageFileSize,
            "imageMimeType": self.imageMimeType,
            "usage": self.usage,
            "useFor": self.useFor,
            "owner": self.owner,
            "aacid": self.aacid,
            "width": self.width,
            "height": self.height,
            "imageUrl": self.imageUrl
        })
    @staticmethod
    def from_dict(obj: Any) -> 'AssetNative':
        cls_object, keys = get_object_model_class(obj, AssetNative, None);
        return AssetNative(cls_object)

